//这段代码的主要功能是实现通过串口与MyCobot机械臂之间的通信，处理来自客户端的数据并将其发送到Atom模块，同时将Atom模块返回的数据发送给客户端。
//代码具体分为几个部分：初始化设置、数据传输、数据处理和消息传递。

#include <MyCobotBasic.h>  // 引入MyCobotBasic库
#include <vector>  // 引入STL vector库

#define SET_BASIC_OUT 0xa0  // 定义宏，表示设置基本输出的命令
#define GET_BASIC_IN 0xa1  // 定义宏，表示获取基本输入的命令

byte HEADER = 0xfe;  // 定义头字节
byte FOOTER = 0xfa;  // 定义尾字节

using namespace std;  // 使用标准命名空间
vector<int> client_data;  // 定义存储客户端数据的向量
vector<int> temp_data;  // 定义临时数据向量
vector<int> atom_data;  // 定义存储Atom模块数据的向量
int count = -1, data = -1;  // 定义计数器和数据变量
int t_begin;  // 定义开始时间变量

void setup() {
    Serial.begin(115200);  // 初始化串口通信，波特率为115200
    /* 如果使用其他开发板，请检查官方文档以确定串口是否支持Serial1。
     * 例如，m5的串口是Serial1（所以将以下Serial1替换为Serial1）*/
    Serial1.begin(1000000);  // 初始化串口1通信，波特率为1000000
    delay(100);  // 延迟100毫秒

    MyCobotBasic myCobot;  // 创建MyCobotBasic对象
    delay(500);  // 延迟500毫秒

    myCobot.powerOn();  // 为MyCobot上电
    delay(500);  // 延迟500毫秒

    pinMode(15, OUTPUT); // 将引脚15设置为输出
    pinMode(5, OUTPUT); // 将引脚5设置为输出
    pinMode(2, OUTPUT); // 将引脚2设置为输出
}

void loop() {
    ClientDataToAtom();  // 将客户端数据发送到Atom模块
    AtomDataToClient();  // 将Atom模块数据发送到客户端
    OtherMsgToClient();  // 处理其他需要发送到客户端的消息
    ClearData();  // 清理数据
}

/*
 * 函数：检查串口WiFi或蓝牙返回消息的头是否为fe fe
 */
bool checkHeader(vector<int> &v_data) {
    if (v_data[0] == HEADER && v_data[1] == HEADER)
        return true;
    return false;
}

/*
 * 函数：处理粘包问题
 */
bool HandleStickyPackets(vector<int> &v_data) {
    // 获取协议长度以确定终止符fa
    vector<int>::iterator it_vdata = v_data.begin() + 2 + v_data[2];
    int len = v_data.size();
    // 首先判断长度是否小于5，指令长度至少为5 --> 处理数据
    if (len >= 5 && len <= (v_data[2] + 3)) {
        if (*it_vdata == FOOTER) {
            v_data.erase(it_vdata + 1, v_data.end());
            return true;
        }
    } 
    return false;
}

/*
 * 函数：处理Atom不处理的一些客户端消息
 */
bool HandleOtherMsg(vector<int> &v_data) {
    bool flag = false;
    switch (v_data[2]) {
        case 4:
            switch (v_data[3]) {
                case SET_BASIC_OUT: {
                    // 0xfe 0xfe 0x04 0xa0 pin_no pin_data 0xfa
                    byte pin_no = v_data[4];
                    pinMode(pin_no, OUTPUT);
                    delay(5);
                    bool pin_data = v_data[5];
                    digitalWrite(pin_no, pin_data);
                }
                break;
            }
            break;
        case 3:
            switch (v_data[3]) {
                case GET_BASIC_IN: {
                    byte pin_no = v_data[4];
                    pinMode(pin_no, INPUT);
                    delay(5);
                    bool pin_state = digitalRead(pin_no);
                    delay(5);
                    v_data[2] = 0x04;
                    v_data.insert(v_data.end() - 1, pin_state);
                    flag = true;
                }
                break;
            }
            break;
    }
    return flag;
}

/*
 * 函数：将Atom返回的消息发送给用户
 */
void SendDataToUser(vector<int> &v_data) {
    if (checkHeader(v_data) && HandleStickyPackets(v_data)) {
        if (HandleOtherMsg(v_data)) {
            for (int i = 0; i < v_data.size(); i++) {
                Serial.write(v_data[i]);
            }
        }
        v_data.clear();
    }
    return;
}

/*
 * 函数：将客户端数据发送到Atom模块
 */
void ClientDataToAtom() {
    if (Serial.available() > 0) {
      count = 1; 
      while (Serial.available() > 0) {
          data = Serial.read();
          Serial1.write(data);
          client_data.push_back(data);
      }
      t_begin = millis();
    }
}

/*
 * 函数：将Atom模块数据发送到客户端
 */
void AtomDataToClient() {
    int temp;
    bool flag = false;
    if (Serial1.available() > 0) {
        count = 0;
        while (Serial1.available() > 0) {
            data = Serial1.read();
            if (data == HEADER && (temp = Serial1.read()) == HEADER) {
                Serial.write(temp);
                flag = true;
            }
            if (flag) {
                Serial.write(data);
            }
        }
    }
}

/*
 * 函数：处理需要发送到客户端的其他消息
 */
void OtherMsgToClient() {
    if (count == 1 && ((millis() - t_begin) > 10)) {
        count = 0;
        for (int i = 0; i < client_data.size(); i++) {
            temp_data.push_back(client_data[i]);
        }
        SendDataToUser(temp_data); 
    }
}

/*
 * 函数：清理数据
 */
void ClearData() {
    if (count == 0){
        client_data.clear();
    }
}
